# Standard library imports
from logging import getLogger
from typing import Any, Self

# Third party imports
from pydantic import (
  BaseModel,
  ConfigDict,
  ModelWrapValidatorHandler,
  RootModel,
  SerializationInfo,
  SerializerFunctionWrapHandler,
  ValidationError,
  ValidationInfo,
  ValidatorFunctionWrapHandler,
  field_validator,
  model_serializer,
  model_validator,
)
from pydantic_core import from_json

logger = getLogger(__name__)


PYDANTIC_CONFIG = ConfigDict(
  populate_by_name=True,
  use_enum_values=True,
  validate_default=True,
  validate_assignment=True,
  coerce_numbers_to_str=True,
)


class CustomRootModel[T](RootModel[T]):
  model_config = PYDANTIC_CONFIG
  _dumping_json: bool = False

  # @model_serializer(mode="wrap", when_used="json")
  # def serialize_as_jsonstr(self, nxt: SerializerFunctionWrapHandler):
  #   if not self._dumping_json:
  #     self._dumping_json = True
  #     return self.model_dump_json()
  #   else:
  #     self._dumping_json = False
  #     return nxt(self)

  @model_validator(mode="wrap")
  @classmethod
  def validate_as_jsonstr(cls, data: Any, handler: ValidatorFunctionWrapHandler) -> Self:
    if isinstance(data, str):
      data = from_json(data)
    return handler(data)

  @model_serializer(mode="wrap", when_used="unless-none")
  def return_self(self, nxt: SerializerFunctionWrapHandler, info: SerializationInfo):
    if not info.mode_is_json():
      return self
    if not self._dumping_json:
      self._dumping_json = True
      return self.model_dump_json()
    else:
      self._dumping_json = False
      return nxt(self)


class CustomBaseModel(BaseModel):
  model_config = PYDANTIC_CONFIG

  @field_validator("*", mode="wrap", check_fields=False)
  @classmethod
  def log_failed_field_validations(cls, data: str, handler: ValidatorFunctionWrapHandler, info: ValidationInfo) -> Any:
    results = None

    results = handler(data)

    return data if results is None else results

  @model_validator(mode="wrap")
  @classmethod
  def log_failed_validation(cls, data: Any, handler: ModelWrapValidatorHandler[Self], info: ValidationInfo) -> Self:
    results = None
    try:
      results = handler(data)
    except ValidationError as e:
      if cls.__name__ != "ScheduledOrderDBEntryModel":
        raise

      for err in e.errors():
        if err["type"] != "missing":
          raise
    return results  # type: ignore



# Standard library imports
from datetime import datetime
from pathlib import Path, PurePosixPath
from re import Pattern

# Third party imports
from dateutil.relativedelta import SU, relativedelta
from pydantic import ConfigDict, Field
from pydantic.dataclasses import dataclass

# First party imports
from scheduled_invoice_processor.environment_init_vars import SETTINGS
from scheduled_invoice_processor.typing_custom import CustomerID, StoreNum


class FixRuffTC001: ...


@dataclass
class FileRegisterData(FixRuffTC001):
  __pydantic_config__ = ConfigDict(
    populate_by_name=True,
    use_enum_values=True,
    validate_default=True,
    validate_assignment=True,
    coerce_numbers_to_str=True,
  )

  storenum: StoreNum
  customer_id: CustomerID
  pickup_date: datetime
  dropoff_date: datetime
  file_pattern: Pattern[str]
  _current_week: bool
  _waiting_folder: PurePosixPath
  _local_copy_folder: Path

  file_names: dict[int, str] = Field(default_factory=dict)
  invoice_nums: dict[int, str] = Field(default_factory=dict)
  pickup_success: dict[int, bool] = Field(default_factory=dict)
  preprocess_success: dict[int, bool] = Field(default_factory=dict)
  dropoff_success: dict[int, bool] = Field(default_factory=dict)

  @property
  def current_week(self) -> bool:
    if self._current_week:
      now = datetime.now(SETTINGS.tz)
      window_start = self.pickup_date - relativedelta(weekday=SU(-1), hour=0, minute=0, second=0)
      window_end = self.dropoff_date + relativedelta(weekday=SU(+1), hour=0, minute=0, second=0)

      return window_start <= now < window_end
    return False

  @property
  def remote_file_locs(self) -> dict[int, PurePosixPath]:
    return {idx: self._waiting_folder / name for idx, name in self.file_names.items()}

  @property
  def local_copy_loc(self) -> dict[int, Path]:
    return {idx: self._local_copy_folder / name for idx, name in self.file_names.items()}

  @property
  def stale(self) -> bool:
    return datetime.now(SETTINGS.tz) > (self.dropoff_date + relativedelta(weekday=SU(+1), hour=0, minute=0, second=0))
